"""SCP Open-API Functions.

This module contains a set of rest api functions accessing SCP Open-API.
"""
import base64
import datetime
import hashlib
import hmac
import json
import logging
import os
import requests
import time
from functools import wraps
from typing import Any, Dict, List, Optional
from urllib import parse

CREDENTIALS_PATH = '~/.scp/scp_credential'
API_ENDPOINT = 'https://openapi.samsungsdscloud.com'
TEMP_VM_JSON_PATH = '/tmp/json/tmp_vm_body.json'

logger = logging.getLogger(__name__)


class SCPClientError(Exception):
    pass


class SCPOngoingRequestError(Exception):
    pass


class SCPCreationFailError(Exception):
    pass


class Metadata:
    """Per-cluster metadata file."""

    def __init__(self, path_prefix: str, cluster_name: str) -> None:
        # TODO(ewzeng): Metadata file is not thread safe. This is fine for
        # now since SkyPilot uses a per-cluster lock for ray-related
        # operations. In the future, add a filelock around __getitem__,
        # __setitem__ and refresh.
        self.path = os.path.expanduser(f'{path_prefix}-{cluster_name}')
        # In case parent directory does not exist
        os.makedirs(os.path.dirname(self.path), exist_ok=True)

    def __getitem__(self, instance_id: str) -> Dict[str, Any]:
        assert os.path.exists(self.path), 'Metadata file not found'
        with open(self.path, 'r') as f:
            metadata = json.load(f)
        return metadata.get(instance_id)

    def __setitem__(self, instance_id: str, value: Optional[Dict[str,
                                                                 Any]]) -> None:
        # Read from metadata file
        if os.path.exists(self.path):
            with open(self.path, 'r') as f:
                metadata = json.load(f)
        else:
            metadata = {}
        # Update metadata
        if value is None:
            if instance_id in metadata:
                metadata.pop(instance_id)  # del entry
            if len(metadata) == 0:
                if os.path.exists(self.path):
                    os.remove(self.path)
                return
        else:
            metadata[instance_id] = value
        # Write to metadata file
        with open(self.path, 'w') as f:
            json.dump(metadata, f)

    def refresh(self, instance_ids: List[str]) -> None:
        """Remove all tags for instances not in instance_ids."""
        if not os.path.exists(self.path):
            return
        with open(self.path, 'r') as f:
            metadata = json.load(f)
        for instance_id in list(metadata.keys()):
            if instance_id not in instance_ids:
                del metadata[instance_id]
        if len(metadata) == 0:
            os.remove(self.path)
            return
        with open(self.path, 'w') as f:
            json.dump(metadata, f)

    def keys(self):
        if not os.path.exists(self.path):
            return []
        with open(self.path, 'r') as f:
            metadata = json.load(f)
            return list(metadata.keys())


def raise_scp_error(response: requests.Response) -> None:
    """Raise SCPCloudError if appropriate. """
    status_code = response.status_code
    if status_code == 200 or status_code == 202:
        return
    try:
        resp_json = response.json()
        message = resp_json['message']
    except (KeyError, json.decoder.JSONDecodeError):
        raise SCPClientError(f'Unexpected error. Status code: {status_code}')

    if status_code == 404:
        raise SCPCreationFailError(f'{status_code}: {message}')

    if "There is an ongoing request" in message:
        raise SCPOngoingRequestError(f'{status_code}: {message}')

    raise SCPClientError(f'{status_code}: {message}')


def _singleton(class_):
    instances = {}

    def get_instance(*args, **kwargs):
        if class_ not in instances:
            instances[class_] = class_(*args, **kwargs)
        return instances[class_]

    return get_instance


def _retry(method, max_tries=50, backoff_s=5):

    @wraps(method)
    def method_with_retries(self, *args, **kwargs):
        try_count = 0
        while try_count < max_tries:
            try:
                return method(self, *args, **kwargs)
            except SCPOngoingRequestError:
                logger.warning("Caught a Ongoing Request. Retrying.")
                try_count += 1
                if try_count < max_tries:
                    time.sleep(backoff_s)
                else:
                    raise

    return method_with_retries


@_singleton
class SCPClient:
    """Wrapper functions for SCP Cloud API."""

    def __init__(self) -> None:
        self.credentials = os.path.expanduser(CREDENTIALS_PATH)
        assert os.path.exists(self.credentials), 'Credentials not found'
        with open(self.credentials, 'r') as f:
            lines = [line.strip() for line in f.readlines() if ' = ' in line]
            self._credentials = {
                line.split(' = ')[0]: line.split(' = ')[1] for line in lines
            }
        self.access_key = self._credentials['access_key']
        self.secret_key = self._credentials['secret_key']
        self.project_id = self._credentials['project_id']
        self.client_type = 'OpenApi'
        self.timestamp = ''
        self.signature = ''

        self.headers = {
            'X-Cmp-AccessKey': f'{self.access_key}',
            'X-Cmp-ClientType': f'{self.client_type}',
            'X-Cmp-ProjectId': f'{self.project_id}',
            'X-Cmp-Timestamp': f'{self.timestamp}',
            'X-Cmp-Signature': f'{self.signature}'
        }

    def create_instance(self, instance_config):
        """Launch new instances."""
        url = f'{API_ENDPOINT}/virtual-server/v3/virtual-servers'
        return self._post(url, instance_config)

    @_retry
    def _get(self, url, contents_key='contents'):
        method = 'GET'
        self.set_timestamp()
        self.set_signature(url=url, method=method)

        response = requests.get(url, headers=self.headers)
        raise_scp_error(response)
        if contents_key is not None:
            return response.json().get(contents_key, [])
        else:
            return response.json()

    @_retry
    def _post(self, url, request_body):
        method = 'POST'
        self.set_timestamp()
        self.set_signature(url=url, method=method)

        response = requests.post(url, json=request_body, headers=self.headers)

        raise_scp_error(response)
        return response.json()

    @_retry
    def _delete(self, url, request_body=None):
        method = 'DELETE'
        self.set_timestamp()
        self.set_signature(url=url, method=method)
        if request_body:
            response = requests.delete(url,
                                       json=request_body,
                                       headers=self.headers)

        else:
            response = requests.delete(url, headers=self.headers)
        raise_scp_error(response)
        return response.json()

    def create_security_group(self, zone_id, vpc, sg_name):
        url = f'{API_ENDPOINT}/security-group/v3/security-groups'
        request_body = {
            "securityGroupName": sg_name,
            "serviceZoneId": zone_id,
            "vpcId": vpc,
            "securityGroupDescription": "skypilot sg"
        }
        return self._post(url, request_body)

    def add_security_group_in_rule(self, sg_id):
        url = f'{API_ENDPOINT}/security-group/v2/security-groups/{sg_id}/rules'
        request_body = {
            "ruleDirection": "IN",
            "services": [{
                "serviceType": "TCP",
                "serviceValue": "22"
            }],
            "sourceIpAddresses": ["0.0.0.0/0"],
            "ruleDescription": "skypilot ssh rue"
        }
        return self._post(url, request_body)

    def add_security_group_out_rule(self, sg_id):
        url = f'{API_ENDPOINT}/security-group/v2/security-groups/{sg_id}/rules'
        request_body = {
            "ruleDirection": "OUT",
            "services": [{
                "serviceType": "TCP",
                "serviceValue": "21"
            }, {
                "serviceType": "TCP",
                "serviceValue": "22"
            }, {
                "serviceType": "TCP",
                "serviceValue": "80"
            }, {
                "serviceType": "TCP",
                "serviceValue": "443"
            }],
            "destinationIpAddresses": ["0.0.0.0/0"],
            "ruleDescription": "skypilot out rue"
        }
        return self._post(url, request_body)

    def add_firewall_inbound_rule(self, firewall_id, internal_ip):
        url = f'{API_ENDPOINT}/firewall/v2/firewalls/{firewall_id}/rules'
        request_body = {
            "sourceIpAddresses": ["0.0.0.0/0"],
            "destinationIpAddresses": [internal_ip],
            "services": [{
                "serviceType": "TCP",
                "serviceValue": "22"
            }],
            "ruleDirection": "IN",
            "ruleAction": "ALLOW",
            "isRuleEnabled": True,
            "ruleLocationType": "FIRST",
            "ruleDescription": "description"
        }
        return self._post(url, request_body)

    def add_firewall_outbound_rule(self, firewall_id, internal_ip):
        url = f'{API_ENDPOINT}/firewall/v2/firewalls/{firewall_id}/rules'
        request_body = {
            "sourceIpAddresses": [internal_ip],
            "destinationIpAddresses": ["0.0.0.0/0"],
            "services": [{
                "serviceType": "TCP",
                "serviceValue": "21"
            }, {
                "serviceType": "TCP",
                "serviceValue": "22"
            }, {
                "serviceType": "TCP",
                "serviceValue": "80"
            }, {
                "serviceType": "TCP",
                "serviceValue": "443"
            }],
            "ruleDirection": "OUT",
            "ruleAction": "ALLOW",
            "isRuleEnabled": True,
            "ruleLocationType": "FIRST",
            "ruleDescription": "description"
        }
        return self._post(url, request_body)

    def terminate_instance(self, vm_id):
        url = f'{API_ENDPOINT}/virtual-server/v2/virtual-servers/{vm_id}'
        return self._delete(url)

    def list_instances(self) -> List[dict]:
        """List existing instances."""
        url = f'{API_ENDPOINT}/virtual-server/v2/virtual-servers'
        return self._get(url)

    def list_catalog(self) -> Dict[str, Any]:
        """List offered instances and their availability."""
        response = requests.get(f'{API_ENDPOINT}/instance-types',
                                headers=self.headers)
        raise_scp_error(response)
        return response.json().get('data', [])

    def get_signature(self, method: str, url: str) -> str:
        url_info = parse.urlsplit(url)
        url = f'{url_info.scheme}://{url_info.netloc}{parse.quote(url_info.path)}'
        if url_info.query:
            enc_params = list(
                map(lambda item: (item[0], parse.quote(item[1][0])),
                    parse.parse_qs(url_info.query).items()))
            url = f'{url}?{parse.urlencode(enc_params)}'

        message = method + url + self.timestamp \
                  + self.access_key + self.project_id + self.client_type
        message = bytes(message, 'utf-8')
        secret = bytes(self.secret_key, 'utf-8')
        signature = str(
            base64.b64encode(
                hmac.new(secret, message, digestmod=hashlib.sha256).digest()),
            'utf-8')

        return str(signature)

    def set_timestamp(self) -> None:
        self.timestamp = str(
            int(
                round(
                    datetime.datetime.timestamp(datetime.datetime.now() -
                                                datetime.timedelta(minutes=1)) *
                    1000)))
        self.headers['X-Cmp-Timestamp'] = self.timestamp

    def set_signature(self, method: str, url: str) -> None:

        self.signature = self.get_signature(url=url, method=method)
        self.headers['X-Cmp-Signature'] = self.signature

    def list_nic_details(self, virtual_server_id) -> List[dict]:
        """List existing instances."""
        url = f'{API_ENDPOINT}/virtual-server/v2/virtual-servers/{virtual_server_id}/nics'
        return self._get(url)

    def get_external_ip(self, virtual_server_id, ip):
        nic_details_list = self.list_nic_details(
            virtual_server_id=virtual_server_id)
        for nic_details in nic_details_list:
            if nic_details['ip'] == ip and nic_details['subnetType'] == 'PUBLIC':
                return nic_details['natIp']
        return None

    def list_zones(self) -> List[dict]:
        url = f'{API_ENDPOINT}/project/v3/projects/{self.project_id}/zones'
        return self._get(url)

    def list_products(self, service_zone_id) -> List[dict]:
        url = f'{API_ENDPOINT}/product/v2/zones/{service_zone_id}/products'
        return self._get(url)

    def list_product_groups(self, service_zone_id) -> List[dict]:
        url = f'{API_ENDPOINT}/product/v2/zones/{service_zone_id}/product-groups'
        return self._get(url)

    def list_vpcs(self, service_zone_id) -> List[dict]:
        url = f'{API_ENDPOINT}/vpc/v2/vpcs?serviceZoneId={service_zone_id}'
        return self._get(url)

    def list_subnets(self) -> List[dict]:
        url = f'{API_ENDPOINT}/subnet/v2/subnets?subnetTypes=PUBLIC'
        return self._get(url)

    def del_security_group(self, sg_id):
        url = f'{API_ENDPOINT}/security-group/v2/security-groups/{sg_id}'
        return self._delete(url)

    def del_firwall_rules(self, firewall_id, rule_id_list):
        url = f'{API_ENDPOINT}/firewall/v2/firewalls/{firewall_id}/rules'
        request_body = {"ruleDeletionType": "PARTIAL", "ruleIds": rule_id_list}
        return self._delete(url, request_body=request_body)

    def list_security_groups(self, vpc_id=None, sg_name=None):
        url = f'{API_ENDPOINT}/security-group/v2/security-groups'
        parameter = []
        if vpc_id is not None:
            parameter.append("vpcId=" + vpc_id)
        if sg_name is not None:
            parameter.append("securityGroupName=" + sg_name)
        if len(parameter) > 0:
            url = url + "?" + "&".join(parameter)
        return self._get(url)

    def list_igw(self):
        url = f'{API_ENDPOINT}/internet-gateway/v2/internet-gateways'
        return self._get(url)

    def get_vm_info(self, vm_id):
        url = f'{API_ENDPOINT}/virtual-server/v3/virtual-servers/{vm_id}'
        return self._get(url, contents_key=None)

    def get_firewal_rule_info(self, firewall_id, rule_id):
        url = f'{API_ENDPOINT}/firewall/v2/firewalls/{firewall_id}/rules/{rule_id}'
        return self._get(url, contents_key=None)

    def list_firwalls(self):
        url = f'{API_ENDPOINT}/firewall/v2/firewalls'
        return self._get(url)

    def list_service_zone_names(self):
        url = f'{API_ENDPOINT}/project/v3/projects/{self.project_id}/zones'
        zone_contents = self._get(url)
        return [content['serviceZoneName'] for content in zone_contents]

    def start_instance(self, vm_id):
        url = f'{API_ENDPOINT}/virtual-server/v2/virtual-servers/{vm_id}/start'
        return self._post(url=url, request_body={})

    def stop_instance(self, vm_id):
        url = f'{API_ENDPOINT}/virtual-server/v2/virtual-servers/{vm_id}/stop'
        return self._post(url=url, request_body={})



"""Utils for building pip wheels.

This module is used for building the wheel for SkyPilot under `~/.sky/wheels`,
and the wheel will be used for installing the SkyPilot on the cluster nodes.

The generated folder is like:
    ~/.sky/wheels/<hash_of_wheel>/sky-<version>-py3-none-any.whl

Whenever a new wheel is built, the old ones will be removed.

The ray up yaml templates under sky/templates depend on the naming of the wheel.
"""
import hashlib
import os
import pathlib
import shutil
import subprocess
import tempfile
from typing import Tuple

import filelock
from packaging import version

import sky
from sky import sky_logging
from sky.backends import backend_utils

logger = sky_logging.init_logger(__name__)

# Local wheel path is same as the remote path.
WHEEL_DIR = pathlib.Path(os.path.expanduser(backend_utils.SKY_REMOTE_PATH))
_WHEEL_LOCK_PATH = WHEEL_DIR.parent / '.wheels_lock'
SKY_PACKAGE_PATH = pathlib.Path(sky.__file__).parent.parent / 'sky'

# NOTE: keep the same as setup.py's setuptools.setup(name=..., ...).
_PACKAGE_WHEEL_NAME = 'skypilot'
_WHEEL_PATTERN = (f'{_PACKAGE_WHEEL_NAME}-'
                  f'{version.parse(sky.__version__)}-*.whl')


def _get_latest_wheel_and_remove_all_others() -> pathlib.Path:
    wheel_name = (f'**/{_WHEEL_PATTERN}')
    try:
        latest_wheel = max(WHEEL_DIR.glob(wheel_name), key=os.path.getctime)
    except ValueError:
        raise FileNotFoundError(
            'Could not find built SkyPilot wheels with glob pattern '
            f'{wheel_name} under {WHEEL_DIR!r}') from None

    latest_wheel_dir_name = latest_wheel.parent
    # Cleanup older wheels.
    for f in WHEEL_DIR.iterdir():
        if f != latest_wheel_dir_name:
            if f.is_dir() and not f.is_symlink():
                shutil.rmtree(f, ignore_errors=True)
    return latest_wheel


def _build_sky_wheel():
    """Build a wheel for SkyPilot."""
    with tempfile.TemporaryDirectory() as tmp_dir:
        # prepare files
        tmp_dir = pathlib.Path(tmp_dir)
        (tmp_dir / 'sky').symlink_to(SKY_PACKAGE_PATH, target_is_directory=True)
        setup_files_dir = SKY_PACKAGE_PATH / 'setup_files'
        for f in setup_files_dir.iterdir():
            if f.is_file():
                shutil.copy(str(f), str(tmp_dir))

        # It is important to normalize the path, otherwise 'pip wheel' would
        # treat the directory as a file and generate an empty wheel.
        norm_path = str(tmp_dir) + os.sep
        try:
            # TODO(suquark): For python>=3.7, 'subprocess.run' supports capture
            # of the output.
            subprocess.run([
                'pip3', 'wheel', '--no-deps', norm_path, '--wheel-dir',
                str(tmp_dir)
            ],
                           stdout=subprocess.DEVNULL,
                           stderr=subprocess.PIPE,
                           check=True)
        except subprocess.CalledProcessError as e:
            raise RuntimeError('Fail to build pip wheel for SkyPilot. '
                               f'Error message: {e.stderr.decode()}') from e

        try:
            wheel_path = next(tmp_dir.glob(_WHEEL_PATTERN))
        except StopIteration:
            raise RuntimeError(
                f'Fail to build pip wheel for SkyPilot under {tmp_dir}. '
                'No wheel file is generated.') from None

        # Use a unique temporary dir per wheel hash, because there may be many
        # concurrent 'sky launch' happening.  The path should be stable if the
        # wheel content hash doesn't change.
        with open(wheel_path, 'rb') as f:
            contents = f.read()
        hash_of_latest_wheel = hashlib.md5(contents).hexdigest()

        wheel_dir = WHEEL_DIR / hash_of_latest_wheel
        wheel_dir.mkdir(parents=True, exist_ok=True)
        shutil.move(str(wheel_path), wheel_dir)


def build_sky_wheel() -> Tuple[pathlib.Path, str]:
    """Build a wheel for SkyPilot, or reuse a cached wheel.

    Caller is responsible for removing the wheel.

    Returns:
        A tuple of (wheel path, wheel hash):
        - wheel_path: A temporary path to a directory holding the wheel; path
        is guaranteed unique per wheel content hash.
        - wheel_hash: The wheel content hash.
    """

    def _get_latest_modification_time(path: pathlib.Path) -> float:
        if not path.exists():
            return -1.
        try:
            return max(os.path.getmtime(root) for root, _, _ in os.walk(path))
        except ValueError:
            return -1.

    # This lock prevents that the wheel is updated while being copied.
    # Although the current caller already uses a lock, we still lock it here
    # to guarantee inherent consistency.
    with filelock.FileLock(_WHEEL_LOCK_PATH):  # pylint: disable=E0110
        # This implements a classic "compare, update and clone" consistency
        # protocol. "compare, update and clone" has to be atomic to avoid
        # race conditions.
        last_modification_time = _get_latest_modification_time(SKY_PACKAGE_PATH)
        last_wheel_modification_time = _get_latest_modification_time(WHEEL_DIR)

        # only build wheels if the wheel is outdated
        if last_wheel_modification_time < last_modification_time:
            if not WHEEL_DIR.exists():
                WHEEL_DIR.mkdir(parents=True, exist_ok=True)
            _build_sky_wheel()

        latest_wheel = _get_latest_wheel_and_remove_all_others()

        wheel_hash = latest_wheel.parent.name

        # Use a unique temporary dir per wheel hash, because there may be many
        # concurrent 'sky launch' happening.  The path should be stable if the
        # wheel content hash doesn't change.
        temp_wheel_dir = pathlib.Path(tempfile.gettempdir()) / wheel_hash
        temp_wheel_dir.mkdir(parents=True, exist_ok=True)
        shutil.copy(latest_wheel, temp_wheel_dir)

    return temp_wheel_dir.absolute(), wheel_hash



"""Configuration for service catalog."""

import contextlib
import functools
import threading

_thread_local_config = threading.local()


@contextlib.contextmanager
def _set_use_default_catalog(value: bool):
    old_value = get_use_default_catalog()
    _thread_local_config.use_default_catalog = value
    try:
        yield
    finally:
        _thread_local_config.use_default_catalog = old_value


# Whether the caller requires the catalog to be narrowed down
# to the account-specific catalog (e.g., removing regions not
# enabled for the current account) or just the raw catalog
# fetched from SkyPilot catalog service. The former is used
# for launching clusters, while the latter for commands like
# `show-gpus`.
def get_use_default_catalog() -> bool:
    if not hasattr(_thread_local_config, 'use_default_catalog'):
        # Should not set it globally, as the global assignment
        # will be executed only once if the module is imported
        # in the main thread, and will not be executed in other
        # threads.
        _thread_local_config.use_default_catalog = False
    return _thread_local_config.use_default_catalog


def use_default_catalog(func):
    """Decorator: disable fetching account-specific catalog.

    The account-specific catalog requires the credentials of the
    cloud to be set.
    """

    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        with _set_use_default_catalog(True):
            return func(*args, **kwargs)

    return wrapper



"""holds common utility function/constants to be used by the providers."""

import logging
from pathlib import Path
import time

RAY_RECYCLABLE = "ray-recyclable"


def get_logger(caller_name):
    """
    Configures the logger of this module for console output and file output
    logs of level DEBUG and higher will be directed to file under LOGS_FOLDER.
    logs of level INFO and higher will be directed to console output.
    """
    logger = logging.getLogger(caller_name)
    LOGS_FOLDER = "/tmp/connector_logs/"  # this node_provider's logs location.
    logger.setLevel(logging.DEBUG)

    Path(LOGS_FOLDER).mkdir(parents=True, exist_ok=True)
    logs_path = LOGS_FOLDER + caller_name + time.strftime("%Y-%m-%d--%H-%M-%S")
    # pylint: disable=line-too-long
    file_formatter = logging.Formatter(
        "%(asctime)s %(levelname)-8s %(message)s", datefmt="%Y-%m-%d %H:%M:%S"
    )

    file_handler = logging.FileHandler(logs_path)
    file_handler.setFormatter(file_formatter)
    file_handler.setLevel(logging.DEBUG)

    console_output_handler = logging.StreamHandler()
    console_output_handler.setFormatter(file_formatter)
    console_output_handler.setLevel(logging.INFO)

    logger.addHandler(file_handler)
    logger.addHandler(console_output_handler)
    return logger



"""Module to enable a single SkyPilot key for all VMs in each cloud.

The `setup_<cloud>_authentication` functions will be called on every cluster
provisioning request.

Specifically, after the ray yaml template file `<cloud>-ray.yml.j2` is filled in
with resource specific information, these functions are called with the filled
in ray yaml config as input,
1. Replace the placeholders in the ray yaml file `skypilot:ssh_user` and
   `skypilot:ssh_public_key_content` with the actual username and public key
   content, i.e., `_replace_ssh_info_in_config`.
2. Setup the `authorized_keys` on the remote VM with the public key content,
   by cloud-init or directly using cloud provider's API.

The local machine's public key should not be uploaded to the
`~/.ssh/sky-key.pub` on the remote VM, because it will cause private/public
key pair mismatch when the user tries to launch new VM from that remote VM
using SkyPilot, e.g., the node is used as a spot controller. (Lambda cloud
is an exception, due to the limitation of the cloud provider. See the
comments in setup_lambda_authentication)
"""
import copy
import functools
import os
import re
import socket
import subprocess
import sys
from typing import Any, Dict, Tuple
import uuid

import colorama
from cryptography.hazmat.primitives import serialization
from cryptography.hazmat.primitives.asymmetric import rsa
from cryptography.hazmat.backends import default_backend
import yaml

from sky import clouds
from sky import sky_logging
from sky.adaptors import gcp, ibm
from sky.utils import common_utils
from sky.utils import subprocess_utils
from sky.utils import ux_utils
from sky.skylet.providers.lambda_cloud import lambda_utils

logger = sky_logging.init_logger(__name__)

# TODO: Should tolerate if gcloud is not installed. Also,
# https://pypi.org/project/google-api-python-client/ recommends
# using Cloud Client Libraries for Python, where possible, for new code
# development.

MAX_TRIALS = 64
# TODO(zhwu): Support user specified key pair.
PRIVATE_SSH_KEY_PATH = '~/.ssh/sky-key'
PUBLIC_SSH_KEY_PATH = '~/.ssh/sky-key.pub'


def _generate_rsa_key_pair() -> Tuple[str, str]:
    key = rsa.generate_private_key(backend=default_backend(),
                                   public_exponent=65537,
                                   key_size=2048)

    private_key = key.private_bytes(
        encoding=serialization.Encoding.PEM,
        format=serialization.PrivateFormat.TraditionalOpenSSL,
        encryption_algorithm=serialization.NoEncryption()).decode(
            'utf-8').strip()

    public_key = key.public_key().public_bytes(
        serialization.Encoding.OpenSSH,
        serialization.PublicFormat.OpenSSH).decode('utf-8').strip()

    return public_key, private_key


def _save_key_pair(private_key_path: str, public_key_path: str,
                   private_key: str, public_key: str) -> None:
    private_key_dir = os.path.dirname(private_key_path)
    os.makedirs(private_key_dir, exist_ok=True)

    with open(
            private_key_path,
            'w',
            opener=functools.partial(os.open, mode=0o600),
    ) as f:
        f.write(private_key)

    with open(public_key_path, 'w') as f:
        f.write(public_key)


def get_or_generate_keys() -> Tuple[str, str]:
    """Returns the aboslute private and public key paths."""
    private_key_path = os.path.expanduser(PRIVATE_SSH_KEY_PATH)
    public_key_path = os.path.expanduser(PUBLIC_SSH_KEY_PATH)
    if not os.path.exists(private_key_path):
        public_key, private_key = _generate_rsa_key_pair()
        _save_key_pair(private_key_path, public_key_path, private_key,
                       public_key)
    else:
        # FIXME(skypilot): ran into failing this assert once, but forgot the
        # reproduction (has private key; but has not generated public key).
        #   AssertionError: /home/ubuntu/.ssh/sky-key.pub
        assert os.path.exists(public_key_path), (
            'Private key found, but associated public key '
            f'{public_key_path} does not exist.')
    return private_key_path, public_key_path


def _replace_ssh_info_in_config(config: Dict[str, Any],
                                public_key: str) -> Dict[str, Any]:
    config_str = common_utils.dump_yaml_str(config)
    config_str = config_str.replace('skypilot:ssh_user',
                                    config['auth']['ssh_user'])
    config_str = config_str.replace('skypilot:ssh_public_key_content',
                                    public_key)
    config = yaml.safe_load(config_str)
    return config


def setup_aws_authentication(config: Dict[str, Any]) -> Dict[str, Any]:
    _, public_key_path = get_or_generate_keys()
    with open(public_key_path, 'r') as f:
        public_key = f.read().strip()
    config = _replace_ssh_info_in_config(config, public_key)
    return config


# Snippets of code inspired from
# https://github.com/ray-project/ray/blob/master/python/ray/autoscaler/_private/gcp/config.py
# Takes in config, a yaml dict and outputs a postprocessed dict
# TODO(weilin): refactor the implementation to incorporate Ray autoscaler to
# avoid duplicated codes.
# Retry for the GCP as sometimes there will be connection reset by peer error.
@common_utils.retry
def setup_gcp_authentication(config: Dict[str, Any]) -> Dict[str, Any]:
    _, public_key_path = get_or_generate_keys()
    with open(public_key_path, 'r') as f:
        public_key = f.read().strip()
    config = copy.deepcopy(config)

    project_id = config['provider']['project_id']
    compute = gcp.build('compute',
                        'v1',
                        credentials=None,
                        cache_discovery=False)

    try:
        project = compute.projects().get(project=project_id).execute()
    except gcp.http_error_exception() as e:
        # Can happen for a new project where Compute Engine API is disabled.
        #
        # Example message:
        # 'Compute Engine API has not been used in project 123456 before
        # or it is disabled. Enable it by visiting
        # https://console.developers.google.com/apis/api/compute.googleapis.com/overview?project=123456
        # then retry. If you enabled this API recently, wait a few minutes for
        # the action to propagate to our systems and retry.'
        if ' API has not been used in project' in e.reason:
            match = re.fullmatch(r'(.+)(https://.*project=\d+) (.+)', e.reason)
            if match is None:
                raise  # This should not happen.
            yellow = colorama.Fore.YELLOW
            reset = colorama.Style.RESET_ALL
            bright = colorama.Style.BRIGHT
            dim = colorama.Style.DIM
            logger.error(
                f'{yellow}Certain GCP APIs are disabled for the GCP project '
                f'{project_id}.{reset}')
            logger.error('Details:')
            logger.error(f'{dim}{match.group(1)}{reset}\n'
                         f'{dim}    {match.group(2)}{reset}\n'
                         f'{dim}{match.group(3)}{reset}')
            logger.error(
                f'{yellow}To fix, enable these APIs by running:{reset} '
                f'{bright}sky check{reset}')
            sys.exit(1)
        else:
            raise
    except socket.timeout:
        logger.error('Socket timed out when trying to get the GCP project. '
                     'Please check your network connection.')
        raise

    project_oslogin: str = next(  # type: ignore
        (item for item in project['commonInstanceMetadata'].get('items', [])
         if item['key'] == 'enable-oslogin'), {}).get('value', 'False')

    if project_oslogin.lower() == 'true':
        logger.info(
            f'OS Login is enabled for GCP project {project_id}. Running '
            'additional authentication steps.')

        # Try to get the os-login user from `gcloud`, as this is the most
        # accurate way to figure out how this gcp user is meant to log in.
        proc = subprocess.run(
            'gcloud compute os-login describe-profile --format yaml',
            shell=True,
            stdout=subprocess.PIPE,
            check=False)
        os_login_username = None
        if proc.returncode == 0:
            try:
                profile = yaml.safe_load(proc.stdout)
                username = profile['posixAccounts'][0]['username']
                if username:
                    os_login_username = username
            except Exception as e:  # pylint: disable=broad-except
                logger.debug('Failed to parse gcloud os-login profile.\n'
                             f'{common_utils.format_exception(e)}')
                pass

        if os_login_username is None:
            # As a fallback, read the account information from the credential
            # file. This works most of the time, but fails if the user's
            # os-login username is not a straightforward translation of their
            # email address, for example because their email address changed
            # within their google workspace after the os-login credentials
            # were established.
            config_path = os.path.expanduser(clouds.gcp.GCP_CONFIG_PATH)
            sky_backup_config_path = os.path.expanduser(
                clouds.gcp.GCP_CONFIG_SKY_BACKUP_PATH)
            assert os.path.exists(sky_backup_config_path), (
                'GCP credential backup file '
                f'{sky_backup_config_path!r} does not exist.')

            with open(sky_backup_config_path, 'r') as infile:
                for line in infile:
                    if line.startswith('account'):
                        account = line.split('=')[1].strip()
                        break
                else:
                    with ux_utils.print_exception_no_traceback():
                        raise RuntimeError(
                            'GCP authentication failed, as the oslogin is '
                            f'enabled but the file {config_path} does not '
                            'contain the account information.')
            os_login_username = account.replace('@', '_').replace('.', '_')
        config['auth']['ssh_user'] = os_login_username

        # Add ssh key to GCP with oslogin
        subprocess.run(
            'gcloud compute os-login ssh-keys add '
            f'--key-file={public_key_path}',
            check=True,
            shell=True,
            stdout=subprocess.DEVNULL)
        # Enable ssh port for all the instances
        enable_ssh_cmd = ('gcloud compute firewall-rules create '
                          'allow-ssh-ingress-from-iap '
                          '--direction=INGRESS '
                          '--action=allow '
                          '--rules=tcp:22 '
                          '--source-ranges=0.0.0.0/0')
        proc = subprocess.run(enable_ssh_cmd,
                              check=False,
                              shell=True,
                              stdout=subprocess.DEVNULL,
                              stderr=subprocess.PIPE)
        if proc.returncode != 0 and 'already exists' not in proc.stderr.decode(
                'utf-8'):
            subprocess_utils.handle_returncode(proc.returncode, enable_ssh_cmd,
                                               'Failed to enable ssh port.',
                                               proc.stderr.decode('utf-8'))
    return _replace_ssh_info_in_config(config, public_key)


def setup_azure_authentication(config: Dict[str, Any]) -> Dict[str, Any]:
    _, public_key_path = get_or_generate_keys()
    with open(public_key_path, 'r') as f:
        public_key = f.read().strip()
    return _replace_ssh_info_in_config(config, public_key)


def setup_lambda_authentication(config: Dict[str, Any]) -> Dict[str, Any]:
    get_or_generate_keys()

    # Ensure ssh key is registered with Lambda Cloud
    lambda_client = lambda_utils.LambdaCloudClient()
    public_key_path = os.path.expanduser(PUBLIC_SSH_KEY_PATH)
    with open(public_key_path, 'r') as f:
        public_key = f.read().strip()
    prefix = f'sky-key-{common_utils.get_user_hash()}'
    name, exists = lambda_client.get_unique_ssh_key_name(prefix, public_key)
    if not exists:
        lambda_client.register_ssh_key(name, public_key)

    # Need to use ~ relative path because Ray uses the same
    # path for finding the public key path on both local and head node.
    config['auth']['ssh_public_key'] = PUBLIC_SSH_KEY_PATH

    # TODO(zhwu): we need to avoid uploading the public ssh key to the
    # nodes, as that will cause problem when the node is used as spot
    # controller, i.e., the public and private key on the node may
    # not match.
    file_mounts = config['file_mounts']
    file_mounts[PUBLIC_SSH_KEY_PATH] = PUBLIC_SSH_KEY_PATH
    config['file_mounts'] = file_mounts

    return config


def setup_ibm_authentication(config):
    """ registers keys if they do not exist in sky folder
    and updates config file.
    keys default location: '~/.ssh/sky-key' and '~/.ssh/sky-key.pub'
    """

    def _get_unique_key_name():
        suffix_len = 10
        return f'skypilot-key-{str(uuid.uuid4())[:suffix_len]}'

    client = ibm.client(region=config['provider']['region'])
    resource_group_id = config['provider']['resource_group_id']

    _, public_key_path = get_or_generate_keys()
    with open(os.path.abspath(os.path.expanduser(public_key_path)),
              'r',
              encoding='utf-8') as file:
        ssh_key_data = file.read().strip()
    # pylint: disable=E1136
    try:
        res = client.create_key(public_key=ssh_key_data,
                                name=_get_unique_key_name(),
                                resource_group={
                                    'id': resource_group_id
                                },
                                type='rsa').get_result()
        vpc_key_id = res['id']
        logger.debug(f'Created new key: {res["name"]}')

    except ibm.ibm_cloud_sdk_core.ApiException as e:
        if 'Key with fingerprint already exists' in e.message:
            for key in client.list_keys().result['keys']:
                if (ssh_key_data in key['public_key'] or
                        key['public_key'] in ssh_key_data):
                    vpc_key_id = key['id']
                    logger.debug(f'Reusing key:{key["name"]}, '
                                 f'matching existing public key.')
                    break
        elif 'Key with name already exists' in e.message:
            raise Exception("""a key with chosen name
                already registered in the specified region""") from e
        else:
            raise Exception('Failed to register a key') from e

    config['auth']['ssh_private_key'] = PRIVATE_SSH_KEY_PATH

    for node_type in config['available_node_types']:
        config['available_node_types'][node_type]['node_config'][
            'key_id'] = vpc_key_id

    # Add public key path to file mounts
    file_mounts = config['file_mounts']
    file_mounts[PUBLIC_SSH_KEY_PATH] = PUBLIC_SSH_KEY_PATH
    config['file_mounts'] = file_mounts

    return config


# Apr, 2023 by Hysun(hysun.he@oracle.com): Added support for OCI
def setup_oci_authentication(config: Dict[str, Any]) -> Dict[str, Any]:
    _, public_key_path = get_or_generate_keys()
    with open(public_key_path, 'r') as f:
        public_key = f.read().strip()

    return _replace_ssh_info_in_config(config, public_key)


def setup_scp_authentication(config: Dict[str, Any]) -> Dict[str, Any]:
    _, public_key_path = get_or_generate_keys()
    with open(public_key_path, 'r') as f:
        public_key = f.read().strip()
    return _replace_ssh_info_in_config(config, public_key)


def setup_kubernetes_authentication(config: Dict[str, Any]) -> Dict[str, Any]:
    get_or_generate_keys()

    # Run kubectl command to add the public key to the cluster.
    public_key_path = os.path.expanduser(PUBLIC_SSH_KEY_PATH)
    key_label = clouds.Kubernetes.SKY_SSH_KEY_SECRET_NAME
    cmd = f'kubectl create secret generic {key_label} ' \
          f'--from-file=ssh-publickey={public_key_path}'
    try:
        subprocess.check_output(cmd, stderr=subprocess.STDOUT, shell=True)
    except subprocess.CalledProcessError as e:
        output = e.output.decode('utf-8')
        suffix = f'\nError message: {output}'
        if 'already exists' in output:
            logger.debug(
                f'Key {key_label} already exists in the cluster, using it...')
        elif any(err in output for err in ['connection refused', 'timeout']):
            with ux_utils.print_exception_no_traceback():
                raise ConnectionError(
                    'Failed to connect to the cluster. Check if your '
                    'cluster is running, your kubeconfig is correct '
                    'and you can connect to it using: '
                    f'kubectl get namespaces.{suffix}') from e
        else:
            logger.error(suffix)
            raise

    return config



import os
import subprocess
import sys
import tempfile
import textwrap
from typing import List, Optional, Tuple, NamedTuple

import colorama
from click import testing as cli_testing
import pytest
import yaml

from sky import cli
from sky.utils import command_runner
from sky.utils import subprocess_utils


# TODO(mluo): Refactor smoke test methods
class Test(NamedTuple):
    name: str
    # Each command is executed serially.  If any failed, the remaining commands
    # are not run and the test is treated as failed.
    commands: List[str]
    teardown: Optional[str] = None
    # Timeout for each command in seconds.
    timeout: int = 15 * 60

    def echo(self, message: str):
        # pytest's xdist plugin captures stdout; print to stderr so that the
        # logs are streaming while the tests are running.
        prefix = f'[{self.name}]'
        message = f'{prefix} {message}'
        message = message.replace('\n', f'\n{prefix} ')
        print(message, file=sys.stderr, flush=True)


def run_one_test(test: Test) -> Tuple[int, str, str]:
    log_file = tempfile.NamedTemporaryFile('a',
                                           prefix=f'{test.name}-',
                                           suffix='.log',
                                           delete=False)
    test.echo(f'Test started. Log: less {log_file.name}')
    for command in test.commands:
        log_file.write(f'+ {command}\n')
        log_file.flush()
        proc = subprocess.Popen(
            command,
            stdout=log_file,
            stderr=subprocess.STDOUT,
            shell=True,
        )
        try:
            proc.wait(timeout=test.timeout)
        except subprocess.TimeoutExpired as e:
            log_file.flush()
            test.echo(e)
            proc.returncode = 1  # None if we don't set it.
            break

        if proc.returncode:
            break

    style = colorama.Style
    fore = colorama.Fore
    outcome = (f'{fore.RED}Failed{style.RESET_ALL}'
               if proc.returncode else f'{fore.GREEN}Passed{style.RESET_ALL}')
    reason = f'\nReason: {command}' if proc.returncode else ''
    msg = (f'{outcome}.'
           f'{reason}'
           f'\nLog: less {log_file.name}\n')
    test.echo(msg)
    log_file.write(msg)
    if proc.returncode == 0 and test.teardown is not None:
        subprocess_utils.run(
            test.teardown,
            stdout=log_file,
            stderr=subprocess.STDOUT,
            timeout=10 * 60,  # 10 mins
            shell=True,
        )

    if proc.returncode:
        raise Exception(f'test failed: less {log_file.name}')


# ---------- Testing On-prem ----------
class TestOnprem:
    """On-prem tests.

    Running these tests requires one or more nodes manually setup to run onprem Sky.
    You can either use a pre-setup cluster the sky team maintains
    (message Michael Luo for ip and credentials) or:
    1. Provision a VM on any cloud provider
    2. Add IP and path to private key in fixtures `_server_ips` and `_private_key` below.
    3. Create a user ubuntu, if not already exists and add the ssh key to allowed keys.
    4. Install Ray and Sky (described in sky onprem readme)
    5. Run tests with `pytest -s -q --tb=short --disable-warnings
       tests/test_onprem.py::TestOnprem`
    """
    # TODO(mluo): Automate on-prem node provisioning using boto3/docker.
    @pytest.fixture
    def server_ips(self):
        yield ['3.142.96.58']

    @pytest.fixture
    def local_cluster_name(self):
        yield 'on-prem-smoke-test'

    @pytest.fixture
    def admin_ssh_user(self):
        yield 'ubuntu'

    @pytest.fixture
    def first_ssh_user(self):
        yield 'test'

    @pytest.fixture
    def second_ssh_user(self):
        yield 'test1'

    @pytest.fixture
    def ssh_private_key(self):
        yield '~/.ssh/on-prem-smoke-key'

    @pytest.fixture
    def admin_cluster_config(self, server_ips, admin_ssh_user, ssh_private_key,
                             local_cluster_name):
        # Generates a cluster config for the sys admin.
        yaml_dict = {
            'cluster': {
                'ips': server_ips,
                'name': local_cluster_name
            },
            'auth': {
                'ssh_user': admin_ssh_user,
                'ssh_private_key': ssh_private_key
            }
        }
        return yaml_dict

    @pytest.fixture
    def cluster_config_setup(self, admin_cluster_config, local_cluster_name):
        # Returns a function that setups up the cluster for the user
        # Set up user's cluster config to be stored in `~/.sky/local/...`.
        def _local_cluster_setup(ssh_user, cluster_name=local_cluster_name):
            local_sky_folder = os.path.expanduser('~/.sky/local/')
            os.makedirs(local_sky_folder, exist_ok=True)
            with open(f'{local_sky_folder}/{cluster_name}.yml', 'w') as f:
                # Change from normal admin to user (emulates admin sending users
                # the private key)
                admin_cluster_config['auth']['ssh_user'] = ssh_user
                admin_cluster_config['cluster']['name'] = cluster_name
                yaml.dump(admin_cluster_config, f)

        return _local_cluster_setup

    @pytest.fixture
    def jobs_reset(self, admin_cluster_config, first_ssh_user, second_ssh_user):
        # Deletes all prior jobs on the cluster.
        head_ip = admin_cluster_config['cluster']['ips'][0]
        ssh_user = admin_cluster_config['auth']['ssh_user']
        ssh_key = admin_cluster_config['auth']['ssh_private_key']

        ssh_credentials = (ssh_user, ssh_key, 'sky-admin-deploy')
        runner = command_runner.SSHCommandRunner(head_ip, *ssh_credentials)

        # Removing /tmp/ray and ~/.sky to reset jobs id to index 0.
        rc = runner.run(('sudo rm -rf /tmp/ray; '
                         f'sudo rm -rf /home/{first_ssh_user}/.sky; '
                         f'sudo rm -rf /home/{second_ssh_user}/.sky'),
                        stream_logs=False)

    @pytest.fixture
    def admin_setup(self, jobs_reset, admin_cluster_config):
        # Sets up Ray cluster on the onprem node.
        cluster_name = admin_cluster_config['cluster']['name']
        with tempfile.NamedTemporaryFile(suffix='.yaml', mode='w') as f:
            yaml.dump(admin_cluster_config, f)
            file_path = f.name
            # Must tear down local cluster in ~/.sky. Running `sky down`
            # removes the local cluster from `sky status` and terminates
            # runnng jobs.
            subprocess.check_output(
                f'sky down -y -p {cluster_name}; sky admin deploy {file_path}',
                shell=True)

    def test_onprem_admin_deploy(self, admin_cluster_config, jobs_reset):
        # Tests running admin setup commands on on-prem cluster.
        name = admin_cluster_config['cluster']['name']
        with tempfile.NamedTemporaryFile(suffix='.yaml', mode='w') as f:
            f.write(str(admin_cluster_config))
            f.flush()
            file_path = f.name
            test = Test(
                'test-admin-deploy',
                [
                    f'sky admin deploy {file_path}',
                ],
                # Cleaning up artifacts created from the test.
                f'rm -rf ~/.sky/local/{name}.yml')
            run_one_test(test)

    def test_onprem_inline(self, local_cluster_name, admin_setup,
                           cluster_config_setup, first_ssh_user):

        # Setups up local cluster config for user 'test'
        cluster_config_setup(first_ssh_user)

        # Tests running commands on on-prem cluster.
        name = local_cluster_name

        test = Test(
            'test_onprem_inline_commands',
            [
                f'sky launch -c {name} -y --env TEST_ENV="hello world" -- "([[ ! -z \\"\$TEST_ENV\\" ]] && [[ ! -z \\"\$SKYPILOT_NODE_IPS\\" ]] && [[ ! -z \\"\$SKYPILOT_NODE_RANK\\" ]]) || exit 1"',
                f'sky logs {name} 1 --status',
                f'sky exec {name} --env TEST_ENV2="success" "([[ ! -z \\"\$TEST_ENV2\\" ]] && [[ ! -z \\"\$SKYPILOT_NODE_IPS\\" ]] && [[ ! -z \\"\$SKYPILOT_NODE_RANK\\" ]]) || exit 1"',
                f'sky logs {name} 2 --status',
            ],
            # Cleaning up artifacts created from the test.
            f'sky down -y {name}; rm -f ~/.sky/local/{name}.yml',
        )
        run_one_test(test)

    def test_onprem_yaml(self, local_cluster_name, admin_setup,
                         cluster_config_setup, first_ssh_user):
        # Setups up local cluster config for user 'test'
        cluster_config_setup(first_ssh_user)

        # Tests running Sky task yaml on on-prem cluster.
        name = local_cluster_name
        yaml_dict = {
            'setup': 'echo "Running setup"',
            'run': textwrap.dedent("""\
                    set -e
                    echo $(whoami)
                    pkill -f ray
                    echo NODE ID: $SKYPILOT_NODE_RANK
                    echo NODE IPS: "$SKYPILOT_NODE_IPS"
                    exit 0""")
        }

        with tempfile.NamedTemporaryFile(suffix='.yaml', mode='w') as f:
            yaml.dump(yaml_dict, f)
            file_path = f.name
            test = Test(
                'test_onprem_yaml',
                [
                    f'sky launch -c {name} -y {file_path}',
                    f'sky logs {name} 1 --status',
                    f'sky exec {name} {file_path}',
                    f'sky logs {name} 2 --status',
                ],
                # Cleaning up artifacts created from the test.
                f'sky down -y {name}; rm -f ~/.sky/local/{name}.yml',
            )
            run_one_test(test)

    def test_onprem_job_queue(self, local_cluster_name, admin_setup,
                              cluster_config_setup, first_ssh_user):
        # Setups up local cluster config for user 'test'
        cluster_config_setup(first_ssh_user)

        # Tests running many jobs on the on-prem cluster.
        name = local_cluster_name
        test = Test(
            'test_onprem_job_queue',
            [
                f'sky launch -y -c {name} -- "echo hi"',
                f'sky exec {name} -d -- "echo hi"',
                f'sky exec {name} -d -- "echo hi"',
                f'sky exec {name} -d -- "echo hi"',
                # Call sleep on the 5th job for `sky cancel` to cancel
                # a running/pending job.
                f'sky exec {name} -d -- "sleep 300"',
                f'sky exec {name} -d -- "echo hi"',
                f'sky cancel {name} 5',
                f'sky logs {name} 1',
                f's=$(sky queue {name}); printf "$s"; echo; echo; printf "$s" | grep "^5\\b" | grep CANCELLED',
            ],
            # Cleaning up artifacts created from the test.
            f'sky down -y {name}; rm -f ~/.sky/local/{name}.yml',
        )
        run_one_test(test)

    def test_onprem_multi_tenancy(self, local_cluster_name, admin_setup,
                                  cluster_config_setup, first_ssh_user,
                                  second_ssh_user):
        # Setups up local cluster config for users `test` and `test1`
        first_cluster_name = local_cluster_name
        cluster_config_setup(first_ssh_user)
        second_cluster_name = 'on-prem-smoke-test-1'
        cluster_config_setup(second_ssh_user, second_cluster_name)

        # Tests running many jobs on the on-prem cluster.
        test = Test(
            'test_onprem_multi_tenancy',
            [
                f'sky launch -y -c {first_cluster_name} -- "echo hi"',
                f'sky launch -y -c {second_cluster_name} -- "echo hi"',
                f'sky exec {first_cluster_name} -d -- "sleep 300"',
                f'sky exec {second_cluster_name} -d -- "sleep 300"',
                f'sky cancel {first_cluster_name} 2',
                'sleep 5',
                f's=$(sky queue {first_cluster_name}); printf "$s"; echo; echo; printf "$s" | grep "^2\\b" | grep CANCELLED',
                # User 1 should not cancel user 2's jobs.
                f's=$(sky queue {second_cluster_name}); printf "$s"; echo; echo; printf "$s" | grep "^2\\b" | grep -v CANCELLED',
                f'sky cancel {second_cluster_name} 2',
                f'sleep 5',
                f's=$(sky queue {second_cluster_name}); printf "$s"; echo; echo; printf "$s" | grep "^2\\b" | grep CANCELLED',
                f'sky logs {first_cluster_name} 1',
                f'sky logs {second_cluster_name} 1'
            ],
            # Cleaning up artifacts created from the test.
            (f'sky down -y {first_cluster_name} {second_cluster_name}; '
             f'rm -f ~/.sky/local/{first_cluster_name}.yml; '
             f'rm -f ~/.sky/local/{second_cluster_name}.yml'))
        run_one_test(test)

    def test_onprem_resource_mismatch(self, local_cluster_name, admin_setup,
                                      cluster_config_setup, first_ssh_user):
        # Setups up local cluster config for user 'test'
        cluster_config_setup(first_ssh_user)
        # Tests on-prem cases when users specify resources that are not on
        # the cluster. This test should error out with
        # ResourcesMismatchError.
        name = local_cluster_name
        cli_runner = cli_testing.CliRunner()
        # Setup the cluster handle and get cluster resources for `sky status`
        cli_runner.invoke(cli.launch, ['-c', name, '--', ''])

        def _test_overspecify_gpus(cluster_name):
            result = cli_runner.invoke(
                cli.launch,
                ['-c', cluster_name, '--gpus', 'V100:256', '--', ''])
            assert 'sky.exceptions.ResourcesMismatchError' not in str(
                type(result.exception))

        with pytest.raises(AssertionError) as e:
            _test_overspecify_gpus(name)
        # Cleaning up artifacts created from the test.
        subprocess.check_output(
            f'sky down -p -y {name}; rm -f ~/.sky/local/{name}.yml', shell=True)



"""Utility functions for UX."""
import contextlib
import sys

import rich.console as rich_console

from sky.utils import env_options

console = rich_console.Console()


def console_newline():
    """Print a newline to the console using rich.

    Useful when catching exceptions inside console.status()
    """
    console.print()


@contextlib.contextmanager
def print_exception_no_traceback():
    """A context manager that prints out an exception without traceback.

    Mainly for UX: user-facing errors, e.g., ValueError, should suppress long
    tracebacks.

    If SKYPILOT_DEBUG environment variable is set, this context manager is a
    no-op and the full traceback will be shown.

    Example usage:

        with print_exception_no_traceback():
            if error():
                raise ValueError('...')
    """
    if env_options.Options.SHOW_DEBUG_INFO.get():
        # When SKYPILOT_DEBUG is set, show the full traceback
        yield
    else:
        original_tracelimit = getattr(sys, 'tracebacklimit', 1000)
        sys.tracebacklimit = 0
        yield
        sys.tracebacklimit = original_tracelimit



#
# (C) Copyright IBM Corp. 2021
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
""" module allocating nodes to Ray's cluster.
 used by the Ray framework as an external connector to IBM provider. """

import concurrent.futures as cf
import inspect
import json
import re
import socket
import threading
import time
from pprint import pprint
from pathlib import Path
from uuid import uuid4

from sky.adaptors import ibm
from typing import Any, Dict, List, Optional

from ray.autoscaler._private.cli_logger import cli_logger
from ray.autoscaler._private.util import hash_runtime_conf, hash_launch_conf
from ray.autoscaler.node_provider import NodeProvider
from ray.autoscaler.tags import (
    NODE_KIND_HEAD,
    NODE_KIND_WORKER,
    TAG_RAY_CLUSTER_NAME,
    TAG_RAY_NODE_KIND,
    TAG_RAY_NODE_NAME,
    TAG_RAY_LAUNCH_CONFIG,
    TAG_RAY_RUNTIME_CONFIG,
    TAG_RAY_USER_NODE_TYPE,
    TAG_RAY_NODE_STATUS,
    TAG_RAY_FILE_MOUNTS_CONTENTS,
)
from sky.skylet.providers.ibm.vpc_provider import IBMVPCProvider
from sky.skylet.providers.ibm.utils import get_logger, RAY_RECYCLABLE

logger = get_logger("node_provider_")

INSTANCE_NAME_UUID_LEN = 8
INSTANCE_NAME_MAX_LEN = 64
PENDING_TIMEOUT = 3600  # period before a node that's yet to run is deleted.
VOLUME_TIER_NAME_DEFAULT = "general-purpose"
# identifies resources created by this package.
# these resources are deleted alongside the node.
VPC_TAGS = ".sky-vpc-tags"


def log_in_out(func):
    """Tracing decorator for debugging purposes"""

    def decorated_func(*args, **kwargs):
        name = func.__name__
        logger.debug(
            f"\n\nEnter {name} from {inspect.stack()[0][3]} "
            f"{inspect.stack()[1][3]} {inspect.stack()[2][3]} with args: "
            f"entered with args:\n{pprint(args)} and kwargs {pprint(kwargs)}"
        )
        try:
            result = func(*args, **kwargs)
            logger.debug(
                f"Leave {name} from {inspect.stack()[1][3]} with result "
                f"Func Result:{pprint(result)}\n\n"
            )
        except Exception:
            cli_logger.error(f"Error in {name}")
            raise
        return result

    return decorated_func


class IBMVPCNodeProvider(NodeProvider):
    """Node Provider for IBM VPC

    Currently, instance tagging is implemented using internal cache

    To communicate with head node from outside cluster private network,
    `use_hybrid_ips` set to True. Then, floating (external) ip allocated to
    cluster head node, while worker nodes are provisioned with private ips only.
    """

    # IBM-TODO alter tagging mechanism to support cloud tags instead of
    #  maintaining VM tags locally (file).
    def _load_tags(self):
        """
        if local tags cache (file) exists (cluster is restarting),
            cache is loaded, deleted nodes are filtered away and result
            is dumped to local cache.
        otherwise, remote head node will initialize the in memory
          and local storage tags cache with the head's cluster tags.
        """

        # local tags cache exists from former runs
        if self.tags_file.is_file():
            all_tags = json.loads(self.tags_file.read_text())
            tags = all_tags.get(self.cluster_name, {})

            # filters instances that were deleted since
            # the last time the head node was up
            for instance_id, instance_tags in tags.items():
                try:
                    self.ibm_vpc_client.get_instance(instance_id)
                    self.nodes_tags[instance_id] = instance_tags
                except ibm.ibm_cloud_sdk_core.ApiException as e:
                    cli_logger.warning(instance_id)
                    if e.message == "Instance not found":
                        logger.error(
                            f"cached instance {instance_id} not found, \
                                will be removed from cache"
                        )
            # dump in-memory cache to local cache (file).
            self.set_node_tags(None, None)

        else:
            name = socket.gethostname()  # returns the instance's (VSI) name
            logger.debug(f"Check if {name} is HEAD")
            # check if current VM is the remote head node
            if self._get_node_type(name) == NODE_KIND_HEAD:
                logger.debug(f"{name} is HEAD")
                # pylint: disable=line-too-long
                node = self.ibm_vpc_client.list_instances(name=name).get_result()[
                    "instances"
                ]
                if node:
                    logger.debug(f"{name} is node in vpc")

                    # remote head node will calculate 2 hash values below
                    ray_bootstrap_config = Path.home() / "ray_bootstrap_config.yaml"
                    config = json.loads(ray_bootstrap_config.read_text())
                    (runtime_hash, mounts_contents_hash) = hash_runtime_conf(
                        config["file_mounts"], None, config
                    )

                    # launch_hash is used to verify the head node belongs
                    # to the cluster. e.g. checked when running --restart-only
                    head_node_config = {}
                    head_node_type = config.get("head_node_type")
                    if head_node_type:
                        head_config = config["available_node_types"][head_node_type]
                        head_node_config.update(head_config["node_config"])
                    launch_hash = hash_launch_conf(head_node_config, config["auth"])

                    head_tags = {
                        TAG_RAY_NODE_KIND: NODE_KIND_HEAD,
                        TAG_RAY_NODE_NAME: name,
                        TAG_RAY_NODE_STATUS: "up-to-date",
                        TAG_RAY_CLUSTER_NAME: self.cluster_name,
                        TAG_RAY_USER_NODE_TYPE: config["head_node_type"],
                        TAG_RAY_RUNTIME_CONFIG: runtime_hash,
                        TAG_RAY_LAUNCH_CONFIG: launch_hash,
                        TAG_RAY_FILE_MOUNTS_CONTENTS: mounts_contents_hash,
                    }

                    logger.debug(f"Setting HEAD node tags {head_tags}")
                    self.set_node_tags(node[0]["id"], head_tags)

    def __init__(self, provider_config, cluster_name):
        """
        Args:
            provider_config (dict): containing the provider segment of
                the cluster's config file.initialized as an instance
                variable of the parent class, hence can accessed via self.
            cluster_name(str): value of cluster_name within the
                 cluster's config file.

        """
        NodeProvider.__init__(self, provider_config, cluster_name)

        self.lock = threading.RLock()
        self.tags_file = Path.home() / VPC_TAGS
        self.iam_api_key = provider_config["iam_api_key"]
        self.resource_group_id = provider_config["resource_group_id"]
        self.iam_endpoint = provider_config.get("iam_endpoint")
        self.region = provider_config["region"]
        self.zone = self.provider_config["availability_zone"]
        self.ibm_vpc_client = ibm.client(region=self.region)

        # vpc_tags may contain the following fields:
        # vpc_id, subnet_id, security_group_id
        self.vpc_tags = None
        self.vpc_provider = IBMVPCProvider(
            self.resource_group_id, self.region, self.cluster_name
        )
        # cache of the cluster's nodes in the format: {node_id:node_tags}
        self.nodes_tags = {}
        # Cache of starting/running/pending(below PENDING_TIMEOUT)nodes.
        # in the format{node_id:node_data}.
        self.cached_nodes = {}
        # cache of the nodes created, but not yet tagged and running.
        # in the format {node_id:time_of_creation}.
        self.pending_nodes = {}
        # ids of nodes scheduled for deletion.
        self.deleted_nodes = []

        self._load_tags()

        # if cache_stopped_nodes == true, nodes will be stopped
        #   instead of deleted to better accommodate future rise in demand
        self.cache_stopped_nodes = provider_config.get("cache_stopped_nodes", True)

    def _get_node_type(self, name):
        """returns the type of node with the specified name.

        Args:
            name (str): node's name
        Returns:
            node type (str): NODE_KIND_HEAD or NODE_KIND_WORKER.
            None: return none if node name doesn't belong to cluster's
                name scheme.
        """
        if f"{self.cluster_name}-{NODE_KIND_WORKER}" in name:
            return NODE_KIND_WORKER
        elif f"{self.cluster_name}-{NODE_KIND_HEAD}" in name:
            return NODE_KIND_HEAD

    def _get_nodes_by_tags(self, filters):
        """
        returns list of nodes' instance data who's tags are matching the specified filters.
        updates self.nodes_tags if no filters were specified or
          the only filter is the node's type.

        This function is used by non_terminate_nodes to update self.cached_nodes
          and self.nodes_tags thus abstains from using the cache. moreover,
          using the cache from this function will create a loop on cache miss.
        Args:
            filters(dict):  specified conditions to filter nodes by.
        Returns:
            list of node data for each node that matches filters.
        """

        nodes = []
        found_new_node = False
        # filters specified are a subset of {cluster_name, node_type}
        if {TAG_RAY_NODE_KIND, TAG_RAY_CLUSTER_NAME}.issuperset(set(filters.keys())):
            if self.vpc_tags:
                # if possible, search within the vpc to narrow the search scope
                result = self.ibm_vpc_client.list_instances(
                    vpc_id=self.vpc_tags["vpc_id"]
                ).get_result()
            else:
                result = self.ibm_vpc_client.list_instances().get_result()
            instances = result["instances"]
            # using pagination to acquire instances from all result pages:
            # https://cloud.ibm.com/apidocs/vpc/latest?code=python#api-pagination
            while result.get("next"):
                start = result["next"]["href"].split("start=")[1]
                result = self.ibm_vpc_client.list_instances(start=start).get_result()
                instances.extend(result["instances"])

            for instance in instances:
                kind = self._get_node_type(instance["name"])
                if kind and instance["id"] not in self.deleted_nodes:
                    if not filters or kind == filters[TAG_RAY_NODE_KIND]:
                        nodes.append(instance)
                        with self.lock:
                            node_tags = self.nodes_tags.setdefault(instance["id"], {})
                            if not node_tags:
                                found_new_node = True
                            node_tags.update(
                                {
                                    TAG_RAY_CLUSTER_NAME: self.cluster_name,
                                    TAG_RAY_NODE_KIND: kind,
                                }
                            )
            # update local tags cache
            if found_new_node:
                self.set_node_tags(None, None)

        # if more complex set of filters is specified, there's no use to search
        # the VPC for them, as this implementation stores tags locally instead
        # of tagging the instances. instead we'll search the cache nodes_tags
        else:  # match filters specified
            with self.lock:

                for node_id, node_tags in self.nodes_tags.items():
                    # skipping nodes that don't match all specified filters,
                    # i.e. not all items in filters match those in self.nodes_tags
                    if not all(item in node_tags.items() for item in filters.items()):
                        logger.debug(
                            f"specified filters: {filters} don't match node "
                            f"tags: {node_tags}"
                        )
                        continue

                    # append to result list nodes that do match the filters:
                    try:
                        # not using cache. see function docs for details.
                        nodes.append(self.ibm_vpc_client.get_instance(node_id).result)
                    except ibm.ibm_cloud_sdk_core.ApiException as e:
                        cli_logger.warning(node_id)
                        if "Instance not found" in e.message:
                            logger.error(f"failed to find vsi {node_id}, skipping")
                            continue
                        logger.error(f"failed to find instance {node_id}, raising")
                        raise e

        return nodes

    def non_terminated_nodes(self, tag_filters) -> List[str]:
        """
        returns list of ids of non terminated nodes,
        matching the specified tags. updates the nodes cache.
        NOTE: this function is called periodically by ray,
            a fact utilized to refresh caches (self.cached_nodes and self.node_tags).
        Args:
            tag_filters(dict): fields by which nodes will be filtered.
        """

        # collecting valid nodes that are either starting, running or pending.
        res_nodes = []

        found_nodes = self._get_nodes_by_tags(tag_filters)

        for node in found_nodes:

            # check if node scheduled for delete
            with self.lock:
                if node["id"] in self.deleted_nodes:
                    logger.info(f"{node['id']} scheduled for delete")
                    continue

            # delete failed nodes and skip nodes in other invalid states
            valid_statuses = ["pending", "starting", "running"]
            if node["status"] not in valid_statuses:
                with self.lock:
                    if node["status"] == "failed":
                        # since non_terminated_nodes is called periodically,
                        # prevent access until node is deleted
                        self._delete_node(node["id"], node)
                        logger.error(
                            """Encountered a failed node.
                            Region might be crowded,
                            Consider retrying later."""
                        )
                    else:
                        logger.debug(
                            f"{node['id']} status {node['status']}"
                            f" not in {valid_statuses}, skipping"
                        )
                        continue

            # remove instance hanging in pending state
            with self.lock:
                if node["id"] in self.pending_nodes:
                    if node["status"] != "running":
                        pending_time = time.time() - self.pending_nodes[node["id"]]
                        logger.debug(f"{node['id']} is pending for {pending_time}")
                        if pending_time > PENDING_TIMEOUT:
                            logger.error(
                                f"pending timeout {PENDING_TIMEOUT} reached, "
                                f"deleting instance {node['id']}"
                            )
                            self._delete_node(node["id"], node)
                            # avoid adding the node to cached_nodes
                            continue
                    else:
                        self.pending_nodes.pop(node["id"], None)

            # verify a floating ip is attached to the node
            nic_id = node["network_interfaces"][0]["id"]
            # pylint: disable=line-too-long
            res = self.ibm_vpc_client.list_instance_network_interface_floating_ips(
                node["id"], nic_id
            ).get_result()
            floating_ips = res["floating_ips"]
            if len(floating_ips) == 0:
                # not adding a node that's yet/failed to
                # to get a floating ip provisioned
                continue
            else:
                node["floating_ips"] = floating_ips

            res_nodes.append(node)

        # update cached_nodes
        for node in res_nodes:
            self.cached_nodes[node["id"]] = node

        return [node["id"] for node in res_nodes]

    def is_running(self, node_id) -> bool:
        """returns whether a node is in status running"""
        with self.lock:
            node = self._get_cached_node(node_id)
            logger.debug(
                f"""node: {node_id} is_running?
                {node["status"] == "running"}"""
            )
            return node["status"] == "running"

    def is_terminated(self, node_id) -> bool:
        """returns True if a node is either not recorded or
        not in any valid status."""
        with self.lock:
            try:
                node = self._get_cached_node(node_id)
                logger.debug(
                    f"""node: {node_id} is_terminated?
                    {node["status"] not in
                    ["running", "starting", "pending"]}"""
                )
                return node["status"] not in ["running", "starting", "pending"]
            # pylint: disable=W0703
            except Exception:
                return True

    def node_tags(self, node_id) -> Dict[str, str]:
        """returns tags of specified node id"""

        with self.lock:
            return self.nodes_tags.get(node_id, {})

    def external_ip(self, node_id) -> Optional[str]:
        """returns node's public ip

        Args:
            node_id (str): id of a node within ibm VPC

        Returns:
            str: public ip of specified node
        """

        with self.lock:
            node = self._get_cached_node(node_id)
            fip = node.get("floating_ips")
            if fip:
                return fip[0]["address"]
        return None

    def internal_ip(self, node_id) -> str:
        """returns the worker's node private ip address"""
        node = self._get_cached_node(node_id)

        # if a bug ocurred, or node data was fetched before primary_ip
        # was assigned, refetch node data from cloud.
        try:
            primary_ip = node["network_interfaces"][0].get("primary_ip")["address"]
            if primary_ip is None:
                node = self._get_node(node_id)
        # pylint: disable=W0703
        except Exception:
            node = self._get_node(node_id)

        return node["network_interfaces"][0].get("primary_ip")["address"]

    def set_node_tags(self, node_id, tags) -> None:
        """
        updates local (file) tags cache.
        also updates in memory cache (self.nodes_tags) if node_id and tags
        are specified.

        Args:
            node_id(str): node unique id.
            tags(dict): specified conditions by which nodes will be filtered.
        """

        with self.lock:
            # update in-memory cache
            if node_id and tags:
                node_tags = self.nodes_tags.setdefault(node_id, {})
                node_tags.update(tags)

            # dump in-memory cache to file.
            all_tags = {}
            if self.tags_file.is_file():
                all_tags = json.loads(self.tags_file.read_text())
            all_tags[self.cluster_name] = self.nodes_tags
            self.tags_file.write_text(json.dumps(all_tags))

    def _get_instance_data(self, name):
        """Returns instance (node) information matching the specified name"""

        instances_data = self.ibm_vpc_client.list_instances(name=name).get_result()
        if len(instances_data["instances"]) > 0:
            return instances_data["instances"][0]
        return None

    def _create_instance(self, name, base_config):
        """
        Creates a new VM instance with the specified name,
        based on the provided base_config configuration dictionary
        Args:
            name(str): name of the instance.
            base_config(dict): specific node relevant data.
            node type segment of the cluster's config file,
             e.g. ray_head_default.
        """

        logger.info(f"Creating new VM instance {name}")

        security_group_identity_model = {"id": self.vpc_tags["security_group_id"]}
        subnet_identity_model = {"id": self.vpc_tags["subnet_id"]}
        primary_network_interface = {
            "name": "eth0",
            "subnet": subnet_identity_model,
            "security_groups": [security_group_identity_model],
        }

        boot_volume_profile = {
            "capacity": base_config.get("boot_volume_capacity", 100),
            "name": f"boot-volume-{uuid4().hex[:4]}",
            "profile": {
                "name": base_config.get("volume_tier_name", VOLUME_TIER_NAME_DEFAULT)
            },
        }

        boot_volume_attachment = {
            "delete_volume_on_instance_delete": True,
            "volume": boot_volume_profile,
        }

        key_identity_model = {"id": base_config["key_id"]}
        profile_name = base_config.get("instance_profile_name")
        if not profile_name:
            raise Exception(
                """Failed to detect 'instance_profile_name'
                key in cluster config file"""
            )

        instance_prototype = {}
        instance_prototype["name"] = name
        instance_prototype["keys"] = [key_identity_model]
        instance_prototype["profile"] = {"name": profile_name}
        instance_prototype["resource_group"] = {"id": self.resource_group_id}
        instance_prototype["vpc"] = {"id": self.vpc_tags["vpc_id"]}
        instance_prototype["image"] = {"id": base_config["image_id"]}

        instance_prototype["zone"] = {"name": self.zone}
        instance_prototype["boot_volume_attachment"] = boot_volume_attachment
        # pylint: disable=line-too-long
        instance_prototype["primary_network_interface"] = primary_network_interface

        try:
            with self.lock:
                resp = self.ibm_vpc_client.create_instance(instance_prototype)
        except ibm.ibm_cloud_sdk_core.ApiException as e:
            if e.code == 400 and "already exists" in e.message:
                return self._get_instance_data(name)
            elif e.code == 400 and "over quota" in e.message:
                cli_logger.error(f"Create VM instance {name} failed due to quota limit")
            else:
                cli_logger.error(
                    f"""Create VM instance for {name}
                    failed with status code {str(e.code)}
                    .\nFailed instance prototype:\n"""
                )
                pprint(instance_prototype)
            raise e

        logger.info(f"VM instance {name} created successfully")
        return resp.result

    def _create_floating_ip(self, base_config):
        """returns unbound floating IP address.
        Creates a new ip if none were found in the config file.
        Args:
            base_config(dict): specific node relevant data.
            node type segment of the cluster's config file,
            e.g. ray_head_default.
        """

        if base_config.get("head_ip"):
            # pylint: disable=line-too-long
            for ip in self.ibm_vpc_client.list_floating_ips().get_result()[
                "floating_ips"
            ]:
                if ip["address"] == base_config["head_ip"]:
                    return ip

        floating_ip_name = f"{RAY_RECYCLABLE}-{uuid4().hex[:4]}"
        # create a new floating ip
        logger.debug(f"Creating floating IP {floating_ip_name}")
        floating_ip_prototype = {}
        floating_ip_prototype["name"] = floating_ip_name
        floating_ip_prototype["zone"] = {"name": self.zone}
        floating_ip_prototype["resource_group"] = {"id": self.resource_group_id}
        response = self.ibm_vpc_client.create_floating_ip(floating_ip_prototype)
        floating_ip_data = response.result

        return floating_ip_data

    def _attach_floating_ip(self, instance, fip_data):
        """
        attach a floating ip to the network interface of an instance
        Args:
            instance(dict): extensive data of a node.
            fip_data(dict): floating ip data.
        """

        fip = fip_data["address"]
        fip_id = fip_data["id"]

        logger.debug(
            f"""Attaching floating IP {fip} to
            VM instance {instance["id"]}"""
        )

        # check if floating ip is not attached yet
        inst_p_nic = instance["primary_network_interface"]

        if inst_p_nic["primary_ip"] and inst_p_nic["id"] == fip_id:
            # floating ip already attached. do nothing
            logger.debug(f"Floating IP {fip} already attached to eth0")
        else:
            # attach floating ip
            self.ibm_vpc_client.add_instance_network_interface_floating_ip(
                instance["id"], instance["network_interfaces"][0]["id"], fip_id
            )

    def _stopped_nodes(self, tags):
        """
        returns either stopped head or stopped worker nodes,
            belonging to this cluster

        requires: TAG_RAY_NODE_KIND key in tags.
        Args:
            tags(dict): set of conditions nodes will be filtered by.
        Returns:
            list of stopped nodes, where each item contains node data.
        """

        # pylint: disable=C0206 W0622
        # the following subset 'filter' is a common denominator for all
        # nodes belonging to the cluster.
        filter = {
            TAG_RAY_CLUSTER_NAME: self.cluster_name,
            TAG_RAY_NODE_KIND: tags[TAG_RAY_NODE_KIND],
        }
        nodes = []
        # filter stopped nodes
        with self.lock:
            for node_id in self.nodes_tags:
                try:
                    node_tags = self.nodes_tags[node_id]
                    if all(item in node_tags.items() for item in filter.items()):
                        node = self.ibm_vpc_client.get_instance(node_id).result
                        state = node["status"]
                        if state in ["stopped", "stopping"]:
                            nodes.append(node)
                except Exception as e:
                    cli_logger.warning(
                        f"node: {node_id} from cache: 'nodes_tags' not found in VPC"
                    )
                    if (
                        isinstance(e, ibm.ibm_cloud_sdk_core.ApiException)
                        and "Instance not found" in e.message
                    ):
                        continue
                    raise e
            return nodes

    def _create_node(self, base_config, tags):
        """
        returns dict {instance_id:instance_data} of newly created node.
        updates tags cache.
        creates a node in the following format:
            ray-{cluster_name}-{node_type}-{uuid}

        Args:
            base_config(dict): specific node relevant data.
                node type segment of the cluster's config file,
                e.g. ray_head_default.
            tags(dict): set of conditions nodes will be filtered by.
        """

        name_tag = tags[TAG_RAY_NODE_NAME]
        if len(name_tag) > INSTANCE_NAME_MAX_LEN - INSTANCE_NAME_UUID_LEN - 1:
            logger.error(
                f"""node name: {name_tag} is longer than
                {INSTANCE_NAME_MAX_LEN - INSTANCE_NAME_UUID_LEN - 1}
                characters"""
            )
            raise Exception("Invalid node name: length check failed")

        # IBM VPC VSI pattern requirement
        pattern = re.compile("^([a-z]|[a-z][-a-z0-9]*[a-z0-9])$")
        res = pattern.match(name_tag)

        if not res:
            logger.error(
                f"""node name {name_tag} doesn't match the naming pattern
                 `[a-z]|[a-z][-a-z0-9]*[a-z0-9]` for IBM VPC instance """
            )
            raise Exception(
                "Invalid node name: pattern check for IBM VPC instance failed"
            )

        # append instance name with uuid
        name = "{name_tag}-{uuid}".format(
            name_tag=name_tag, uuid=uuid4().hex[:INSTANCE_NAME_UUID_LEN]
        )

        # create instance in vpc
        instance = self._create_instance(name, base_config)

        # record creation time. used to discover hanging nodes.
        with self.lock:
            self.pending_nodes[instance["id"]] = time.time()

        tags[TAG_RAY_CLUSTER_NAME] = self.cluster_name
        tags[TAG_RAY_NODE_NAME] = name
        self.set_node_tags(instance["id"], tags)

        # create and attach a floating ip to the node
        fip_data = self._create_floating_ip(base_config)
        self._attach_floating_ip(instance, fip_data)

        return {instance["id"]: instance}

    def create_node(self, base_config, tags, count) -> Dict[str, Any]:
        """
        returns dict of {instance_id:instance_data} of nodes.
        creates 'count' number of nodes.
        if enabled in base_config,
        tries to re-run stopped nodes before creation of new instances.

        Args:
            base_config(dict): specific node relevant data.
                node type segment of the cluster's config file,
                e.g. ray_head_default.
                a template shared by all nodes
                when creating multiple nodes (count>1).
            tags(dict): set of conditions nodes will be filtered by.
            count(int): number of nodes to create.

        """

        # Create a VPC when creating a head node
        # or fetch (for any node type) when restarting Ray
        if not self.vpc_tags:
            self.vpc_tags = self.vpc_provider.create_or_fetch_vpc(
                self.region, self.zone
            )

        stopped_nodes_dict = {}
        futures = []

        # Try to reuse previously stopped nodes with compatible configs
        if self.cache_stopped_nodes:
            stopped_nodes = self._stopped_nodes(tags)
            stopped_nodes_ids = [n["id"] for n in stopped_nodes]
            stopped_nodes_dict = {n["id"]: n for n in stopped_nodes}

            if stopped_nodes:
                cli_logger.print(
                    f"Reusing nodes {stopped_nodes_ids}. "
                    "To disable reuse, set `cache_stopped_nodes: False` "
                    "under `provider` in the cluster configuration."
                )

            for node in stopped_nodes:
                logger.info(f"Starting instance {node['id']}")
                self.ibm_vpc_client.create_instance_action(node["id"], "start")
            # added delay due to a race condition.
            # discussed at the final comment of this function.
            time.sleep(1)

            # update tags of stopped nodes
            for node_id in stopped_nodes_ids:
                self.set_node_tags(node_id, tags)

            count -= len(stopped_nodes_ids)

        created_nodes_dict = {}

        # create multiple instances concurrently
        if count:
            with cf.ThreadPoolExecutor(count) as ex:
                for _ in range(count):
                    futures.append(ex.submit(self._create_node, base_config, tags))

            for future in cf.as_completed(futures):
                created_node = future.result()
                created_nodes_dict.update(created_node)

        all_created_nodes = stopped_nodes_dict
        all_created_nodes.update(created_nodes_dict)

        # this sleep is required due to race condition with non_terminated_nodes
        # called in separate thread by autoscaler. not a lost as anyway the vsi
        # operating system takes time to start. can be removed after
        # https://github.com/ray-project/ray/issues/28150 resolved
        time.sleep(5)

        return all_created_nodes

    def _delete_node(self, node_id, node_data=None):
        """deletes a node and removes it from the caches.
        if it's a (not failed) head node, deletes the vpc.

        if node_data is specified uses it instead of fetching it
        by using the node_id.

        Args:
            node_id (str): id of the node.
            node_data (dict): node data received by the api: get_instance()

        Raises:
            ApiException: raises VPC api exception if deletion failed
              for any reason other then "resource not found"
        """

        logger.info("Deleting VM instance {}".format(node_id))
        # avoid fetching data if specified. Skipping this optimization
        # could cause a loop if node is 'failed' since _get_cached_node()
        # would eventually invoke non_terminated_nodes()
        # which in turn will invoke this function.
        node = node_data if node_data else self._get_cached_node(node_id)

        # get vpc_id to delete if deleting head node which isn't a failed node
        # pylint: disable=line-too-long
        vpc_id_to_delete = None
        if (
            self._get_node_type(node["name"]) == NODE_KIND_HEAD
            and node["status"] != "failed"
        ):
            if self.vpc_tags:
                # if it was initialized it's sure to contain `vpc_id`
                vpc_id_to_delete = self.vpc_tags["vpc_id"]
            else:
                vpc_id_to_delete = self.ibm_vpc_client.get_instance(
                    node_id
                ).get_result()["vpc"]["id"]

            # if the request is executed by the remote head like in `sky autostop --down`
            # , in contrast to `sky down`, use cloud functions to delete the VPC
            # with its associated resources.
            hostname = socket.gethostname()  # returns the instance's (VSI) name
            if self._get_node_type(hostname) == NODE_KIND_HEAD:
                self.vpc_provider.remote_cluster_removal(vpc_id_to_delete, self.region)

        try:
            self.ibm_vpc_client.delete_instance(node_id)

            with self.lock:
                # update node deletion in all caches
                self.nodes_tags.pop(node_id, None)
                self.pending_nodes.pop(node_id, None)
                self.deleted_nodes.append(node_id)
                self.cached_nodes.pop(node_id, None)

                # calling set_node_tags with None
                # will dump self.nodes_tags cache to file
                self.set_node_tags(None, None)

            # delete all ips created by this module
            floating_ips = node.get("floating_ips", [])
            for ip in floating_ips:
                if ip["name"].startswith(RAY_RECYCLABLE):
                    self.ibm_vpc_client.delete_floating_ip(ip["id"])

        # catch exceptions from instance/ip deletion
        except ibm.ibm_cloud_sdk_core.ApiException as e:
            # allow 'resource not found' exceptions to pass through
            # suggesting the resource was already removed
            if e.code == 404:
                pass
            else:
                raise e

        # Terminate VPC if node deleted is a head node
        if vpc_id_to_delete:
            if self.poll_instance_deleted(node_id):
                self.vpc_provider.delete_vpc(vpc_id_to_delete, self.region)

    def terminate_nodes(self, node_ids):

        if not node_ids:
            return

        futures = []
        with cf.ThreadPoolExecutor(len(node_ids)) as ex:
            for node_id in node_ids:
                logger.debug(f"NodeProvider Terminating node {node_id}")
                futures.append(ex.submit(self.terminate_node, node_id))

        for future in cf.as_completed(futures):
            future.result()

    @log_in_out
    def terminate_node(self, node_id):
        """Deletes the VM instance and the associated volume.
        if cache_stopped_nodes is set to true in the cluster config file,
        nodes are stopped instead."""

        try:
            if self.cache_stopped_nodes:
                cli_logger.print(
                    f"Stopping instance {node_id}. To terminate instead, "
                    "set `cache_stopped_nodes: False` "
                    "under `provider` in the cluster configuration"
                )

                self.ibm_vpc_client.create_instance_action(node_id, "stop")
            else:
                cli_logger.print(f"Terminating instance {node_id}")
                self._delete_node(node_id)

        except ibm.ibm_cloud_sdk_core.ApiException as e:
            if e.code == 404:
                pass
            else:
                raise e

    def _get_node(self, node_id):
        """Refresh and get info for this node, updating the cache."""
        err_msg = f"failed to get instance with id {node_id}"
        # updates self.cached_nodes and nodes_tags
        self.non_terminated_nodes({})

        if node_id in self.cached_nodes:
            return self.cached_nodes[node_id]
        else:
            logger.error(err_msg)
            raise Exception(err_msg)

    def _get_cached_node(self, node_id):
        """Return node info from cache if possible, otherwise fetches it."""
        if node_id in self.cached_nodes:
            return self.cached_nodes[node_id]

        return self._get_node(node_id)

    def poll_instance_deleted(self, instance_id):
        tries = 20  # waits up to 200sec with 10 sec interval
        sleep_interval = 10
        while tries:
            try:
                self.ibm_vpc_client.get_instance(instance_id).get_result()
            except ibm.ibm_cloud_sdk_core.ApiException:
                logger.debug(f"Deleted VM instance with id: {instance_id}")
                return True

            tries -= 1
            time.sleep(sleep_interval)
        logger.Error("\nFailed to delete instance within expected time frame\n")
        return False

    @staticmethod
    def bootstrap_config(cluster_config) -> Dict[str, Any]:
        return cluster_config
